==> web-ui-pack-0.4.1.tgz: package/controls/number.js <==
import onScroll from "../helpers/onScroll";
import { mathSumFloat, onEvent } from "../indexHelpers";
import localeInfo from "../objects/localeInfo";
import WUPBaseControl from "./baseControl";
import WUPTextControl from "./text";
const tagName = "wup-num";
/** Form-control with number-input
 * @example
  const el = document.createElement("wup-num");
  el.$options.name = "number";
  el.$options.validations = {
    min: 1,
    max: 1000,
  };

  const form = document.body.appendChild(document.createElement("wup-form"));
  form.appendChild(el);
  // or HTML
  <wup-form>
    <wup-num name="number" validations="myValidations"/>
  </wup-form>;
 * @see WUPTextControl
 */
export default class WUPNumberControl extends WUPTextControl {
    #ctr = this.constructor;
    static get observedOptions() {
        const arr = super.observedOptions;
        arr.push("format");
        return arr;
    }
    /** Default options - applied to every element. Change it to configure default behavior */
    static $defaults = {
        ...WUPTextControl.$defaults,
        validationRules: {
            ...WUPBaseControl.$defaults.validationRules,
            _mask: WUPTextControl.$defaults.validationRules._mask,
            // _parse: WUPTextControl.$defaults.validationRules._parse as any,
            min: (v, setV) => (v == null || v < setV) && `Min value is ${setV}`,
            max: (v, setV) => (v == null || v > setV) && `Max value is ${setV}`,
        },
    };
    // @ts-expect-error reason: validationRules is different
    $options = {
        ...this.#ctr.$defaults,
    };
    // @ts-expect-error reason: validationRules is different
    _opts = this.$options;
    /** Returns $options.format joined with defaults */
    get $format() {
        const f = this._opts.format;
        return {
            sepDecimal: localeInfo.sepDecimal,
            sep1000: localeInfo.sep1000,
            ...f,
            maxDecimal: f?.maxDecimal ?? f?.minDecimal ?? 0,
            minDecimal: f?.minDecimal ?? 0,
        };
    }
    // WARN usage format #.### impossible because unclear what sepDec/sep100 and what if user wants only limit decimal part
    /** Called when need to format value */
    valueToInput(v) {
        if (v == null) {
            return "";
        }
        if (this._opts.mask) {
            return v.toString();
        }
        // eslint-disable-next-line prefer-const
        let [int, dec] = v.toString().split(".");
        const f = this.$format;
        if (f.sep1000) {
            for (let i = int.length - 3; i > 0; i -= 3) {
                int = `${int.substring(0, i)}${f.sep1000}${int.substring(i)}`;
            }
        }
        if (dec || f.minDecimal || f.maxDecimal) {
            dec = dec === undefined ? "" : dec.substring(0, Math.min(f.maxDecimal, dec.length));
            dec += "0".repeat(Math.max(f.minDecimal - dec.length, 0));
            if (dec.length) {
                return int + this.$format.sepDecimal + dec;
            }
        }
        return int;
    }
    parse(text) {
        const v = Number(text);
        if (!text || Number.isNaN(v)) {
            return undefined;
        }
        return v;
    }
    canParseInput(text) {
        const f = this.$format;
        if (f.maxDecimal > 0 && text.endsWith(f.sepDecimal)) {
            return false; // case "4.|" - use must able to type '.'
        }
        return true;
    }
    parseInput(text) {
        // such parsing is better then Number.parse because ignores wrong chars
        let v = 0;
        let ok = false;
        let dec = null;
        let decLn = 1;
        let decI = 0;
        const f = this.$format;
        for (let i = 0; i < text.length; ++i) {
            const ascii = text.charCodeAt(i);
            const num = ascii - 48;
            if (num >= 0 && num <= 9) {
                ok = true;
                if (dec === null) {
                    v = v * 10 + num;
                }
                else if (f.maxDecimal >= ++decI) {
                    dec = dec * 10 + num;
                    decLn *= 10;
                }
            }
            else if (ascii === f.sepDecimal.charCodeAt(0)) {
                // WARN: expected sepDecimal is single char
                dec = 0;
            }
        }
        if (dec) {
            v += dec / decLn;
        }
        if (!ok) {
            v = undefined; // case impossible but maybe...
        }
        else if (text.charCodeAt(0) === 45) {
            v *= -1; // case: "-123"
        }
        if (ok && (v > Number.MAX_SAFE_INTEGER || v < Number.MIN_SAFE_INTEGER)) {
            this.declineInput(); // decline looks better than error "Out of range"
            throw new RangeError("Out of range");
            // return v as any;
        }
        if (!this._opts.mask) {
            // otherwise it conflicts with mask
            const next = this.valueToInput(v);
            const hasDeclined = this._canShowDeclined && text.length > next.length;
            if (hasDeclined && next === this.valueToInput(this.$value)) {
                // WARN: don't compare v === this.$value because $value=4.567 but format can be 4.56
                this.declineInput();
            }
            else if (text !== next) {
                const el = this.$refInput;
                // fix cursor position
                const pos = el.selectionStart || 0;
                let dp = 0;
                for (let i = 0, k = 0; k < pos && i < next.length; ++i, ++k) {
                    const ascii = text.charCodeAt(k);
                    if (!(ascii >= 48 && ascii <= 57)) {
                        --i;
                        --dp;
                    }
                    else if (next[i] !== text[k]) {
                        ++dp;
                        --k;
                    }
                }
                // el.value = next;
                this.setInputValue(v); // WARN: it refreshes onceError but calls valueToInput again
                const end = pos + dp;
                el.setSelectionRange(end, end);
            }
        }
        return v;
    }
    canHandleUndo() {
        return true;
    }
    setInputValue(v) {
        const txt = this.valueToInput(v);
        super.setInputValue(txt);
    }
    renderControl() {
        super.renderControl();
        this.$refInput.setAttribute("role", "spinbutton");
    }
    gotChanges(propsChanged) {
        super.gotChanges(propsChanged);
        propsChanged?.includes("format") && this.setInputValue(this.$value);
    }
    gotBeforeInput(e) {
        super.gotBeforeInput(e);
        if (!this._opts.mask) {
            const el = e.target;
            let pos = el.selectionStart || 0;
            if (pos === el.selectionEnd) {
                let isBefore = false;
                switch (e.inputType) {
                    case "deleteContentBackward":
                        --pos;
                        isBefore = true;
                    // eslint-disable-next-line no-fallthrough
                    case "deleteContentForward":
                        if (el.value[pos] === this.$format.sep1000) {
                            pos += isBefore ? 0 : 1; // case "1|,234" + Delete => 1|34
                            el.setSelectionRange(pos, pos);
                        }
                        break;
                    default:
                        break;
                }
            }
        }
    }
    _canShowDeclined;
    gotInput(e) {
        if (!this._opts.mask) {
            switch (e.inputType) {
                case "insertText":
                case "insertFromPaste":
                    this._canShowDeclined = true;
                    break;
                default:
                    delete this._canShowDeclined;
                    break;
            }
        }
        super.gotInput(e);
    }
    gotFocus() {
        const r = super.gotFocus();
        this.$refInput.setAttribute("inputmode", "numeric"); // otherwise textControl removes it if mask isn't applied
        r.push(onScroll(this, //
        (v) => this.gotIncrement(-1 * v), { skip: () => this.$isReadOnly || this.$isDisabled })); // allow inc/dec via scroll/swipe
        r.push(onEvent(this, "keyup", (e) => {
            /* istanbul ignore else */
            if (!e.altKey)
                delete this._isAltDown;
            /* istanbul ignore else */
            if (!e.shiftKey)
                delete this._isShiftDown;
            /* istanbul ignore else */
            if (!e.ctrlKey)
                delete this._isCtrlDown;
        }));
        r.push(() => {
            delete this._isAltDown;
            delete this._isShiftDown;
            delete this._isCtrlDown;
        });
        return r;
    }
    _isAltDown;
    _isShiftDown;
    _isCtrlDown;
    gotKeyDown(e) {
        super.gotKeyDown(e);
        if (e.altKey)
            this._isAltDown = true;
        if (e.shiftKey)
            this._isShiftDown = true;
        if (e.ctrlKey)
            this._isCtrlDown = true;
        let v = 0;
        switch (e.key) {
            case "ArrowUp":
                ++v;
                break;
            case "ArrowDown":
                --v;
                break;
            default:
                break;
        }
        if (v) {
            e.preventDefault();
            this.gotIncrement(v);
        }
    }
    /** Called when user tries to increment/decrement value (via ArrowKeys/Mouse/Swipe) */
    gotIncrement(dval) {
        if (this._isAltDown)
            dval *= 0.1;
        else if (this._isShiftDown)
            dval *= 10;
        else if (this._isCtrlDown)
            dval *= 100;
        const v = +(this.$value ?? 0);
        const hasFloat = this._isAltDown || v % 1 !== 0;
        const next = hasFloat ? mathSumFloat(v, dval) : v + dval;
        const el = this.$refInput;
        const inputType = dval > 0 ? "_inc" : "_dec";
        const data = this.valueToInput(next);
        if (el.dispatchEvent(new InputEvent("beforeinput", { inputType, data, bubbles: true, cancelable: true }))) {
            el.value = data;
            el.dispatchEvent(new InputEvent("input", { inputType, bubbles: true }));
        }
    }
}
customElements.define(tagName, WUPNumberControl);

==> web-ui-pack-0.8.1.tgz: package/controls/number.js <==
import onScroll from "../helpers/onScroll";
import { mathSumFloat, onEvent } from "../indexHelpers";
import localeInfo from "../objects/localeInfo";
import WUPBaseControl from "./baseControl";
import WUPTextControl from "./text";
const tagName = "wup-num";
/** Form-control with number-input
 * @example
  const el = document.createElement("wup-num");
  el.$options.name = "number";
  el.$options.validations = {
    min: 1,
    max: 1000,
  };

  const form = document.body.appendChild(document.createElement("wup-form"));
  form.appendChild(el);
  // or HTML
  <wup-form>
    <wup-num name="number" validations="myValidations"/>
  </wup-form>;
 * @see {@link WUPTextControl} */
export default class WUPNumberControl extends WUPTextControl {
    #ctr = this.constructor;
    static get observedOptions() {
        const arr = super.observedOptions;
        arr.push("format");
        return arr;
    }
    /** Default options - applied to every element. Change it to configure default behavior */
    // @ts-expect-error - min: string & min: number is invalid
    static $defaults = {
        ...WUPTextControl.$defaults,
        validationRules: {
            ...WUPBaseControl.$defaults.validationRules,
            _mask: WUPTextControl.$defaults.validationRules._mask,
            // _parse: WUPTextControl.$defaults.validationRules._parse as any,
            min: (v, setV, c) => (v == null || v < setV) && `Min value is ${c.valueToInput(setV)}`,
            max: (v, setV, c) => (v == null || v > setV) && `Max value is ${c.valueToInput(setV)}`,
        },
    };
    /** Returns $options.format joined with defaults */
    get $format() {
        const f = this._opts.format;
        return {
            sepDecimal: localeInfo.sepDecimal,
            sep1000: localeInfo.sep1000,
            ...f,
            maxDecimal: f?.maxDecimal ?? f?.minDecimal ?? 0,
            minDecimal: f?.minDecimal ?? 0,
        };
    }
    // WARN usage format #.### impossible because unclear what sepDec/sep100 and what if user wants only limit decimal part
    valueToInput(v) {
        if (v == null) {
            return "";
        }
        if (this._opts.mask) {
            return v.toString();
        }
        // eslint-disable-next-line prefer-const
        let [int, dec] = v.toString().split(".");
        const f = this.$format;
        if (f.sep1000) {
            const to = v > 0 ? 0 : 1;
            for (let i = int.length - 3; i > to; i -= 3) {
                int = `${int.substring(0, i)}${f.sep1000}${int.substring(i)}`;
            }
        }
        if (dec || f.minDecimal || f.maxDecimal) {
            dec = dec === undefined ? "" : dec.substring(0, Math.min(f.maxDecimal, dec.length));
            dec += "0".repeat(Math.max(f.minDecimal - dec.length, 0));
            if (dec.length) {
                return int + this.$format.sepDecimal + dec;
            }
        }
        return int;
    }
    parse(text) {
        const v = Number(text);
        if (!text || Number.isNaN(v)) {
            return undefined;
        }
        return v;
    }
    canParseInput(text) {
        const f = this.$format;
        if (f.maxDecimal > 0 && text.endsWith(f.sepDecimal)) {
            return false; // case "4.|" - use must able to type '.'
        }
        return true;
    }
    parseInput(text) {
        // such parsing is better then Number.parse because ignores wrong chars
        let v = 0;
        let ok = false;
        let dec = null;
        let decLn = 1;
        let decI = 0;
        const f = this.$format;
        for (let i = 0; i < text.length; ++i) {
            const ascii = text.charCodeAt(i);
            const num = ascii - 48;
            if (num >= 0 && num <= 9) {
                ok = true;
                if (dec === null) {
                    v = v * 10 + num;
                }
                else if (f.maxDecimal >= ++decI) {
                    dec = dec * 10 + num;
                    decLn *= 10;
                }
            }
            else if (ascii === f.sepDecimal.charCodeAt(0)) {
                // WARN: expected sepDecimal is single char
                dec = 0;
            }
        }
        if (dec) {
            v += dec / decLn;
        }
        if (!ok) {
            v = undefined; // case impossible but maybe...
        }
        else if (text.charCodeAt(0) === 45) {
            v *= -1; // case: "-123"
        }
        if (ok && (v > Number.MAX_SAFE_INTEGER || v < Number.MIN_SAFE_INTEGER)) {
            this.declineInput(); // decline looks better than error "Out of range"
            throw new RangeError("Out of range");
            // return v as any;
        }
        if (!this._opts.mask) {
            // otherwise it conflicts with mask
            const next = this.valueToInput(v);
            const hasDeclined = this._canShowDeclined && text.length > next.length;
            if (hasDeclined && next === this.valueToInput(this.$value)) {
                // WARN: don't compare v === this.$value because $value=4.567 but format can be 4.56
                this.declineInput();
            }
            else if (text !== next) {
                const el = this.$refInput;
                // fix cursor position
                const pos = el.selectionStart || 0;
                let dp = 0;
                for (let i = 0, k = 0; k < pos && i < next.length; ++i, ++k) {
                    const ascii = text.charCodeAt(k);
                    if (!(ascii >= 48 && ascii <= 57)) {
                        --i;
                        --dp;
                    }
                    else if (next[i] !== text[k]) {
                        ++dp;
                        --k;
                    }
                }
                // el.value = next;
                this.setInputValue(v, 3 /* SetValueReasons.userInput */); // WARN: it refreshes onceError but calls valueToInput again
                const end = pos + dp;
                el.setSelectionRange(end, end);
            }
        }
        return v;
    }
    // @ts-expect-error - because expected string
    setInputValue(v, reason) {
        const txt = this.valueToInput(v);
        super.setInputValue(txt, reason);
    }
    renderControl() {
        super.renderControl();
        this.$refInput.setAttribute("role", "spinbutton");
    }
    gotChanges(propsChanged) {
        super.gotChanges(propsChanged);
        if ((!propsChanged && this._opts.format) || propsChanged?.includes("format")) {
            this.setInputValue(this.$value, 3 /* SetValueReasons.userInput */);
        }
    }
    gotBeforeInput(e) {
        super.gotBeforeInput(e);
        if (!this._opts.mask) {
            const el = e.target;
            let pos = el.selectionStart || 0;
            if (pos === el.selectionEnd) {
                let isBefore = false;
                switch (e.inputType) {
                    case "deleteContentBackward":
                        --pos;
                        isBefore = true;
                    // eslint-disable-next-line no-fallthrough
                    case "deleteContentForward":
                        if (el.value[pos] === this.$format.sep1000) {
                            pos += isBefore ? 0 : 1; // case "1|,234" + Delete => 1|34
                            el.setSelectionRange(pos, pos);
                        }
                        break;
                    default:
                        break;
                }
            }
        }
    }
    _canShowDeclined;
    gotInput(e) {
        if (!this._opts.mask) {
            switch (e.inputType) {
                case "insertText":
                case "insertFromPaste":
                    this._canShowDeclined = true;
                    break;
                default:
                    delete this._canShowDeclined;
                    break;
            }
        }
        super.gotInput(e);
    }
    gotFocus(ev) {
        const r = super.gotFocus(ev);
        this.$refInput.setAttribute("inputmode", "numeric"); // otherwise textControl removes it if mask isn't applied
        r.push(onScroll(this, //
        (v) => this.gotIncrement(-1 * v), { skip: () => this.$isReadOnly || this.$isDisabled })); // allow inc/dec via scroll/swipe
        this.style.overflow = ""; // disable inline-style from onScroll helper
        r.push(onEvent(this, "keyup", (e) => {
            if (!e.altKey) {
                delete this._isAltDown;
                e.key === "Alt" && this._wasInc && e.preventDefault(); // otherwise focus moves to browser
                delete this._wasInc;
            }
            if (!e.shiftKey)
                delete this._isShiftDown;
            if (!e.ctrlKey)
                delete this._isCtrlDown;
        }, { passive: false }));
        r.push(() => {
            delete this._wasInc;
            delete this._isAltDown;
            delete this._isShiftDown;
            delete this._isCtrlDown;
        });
        return r;
    }
    /** To prevent focus to browser panel by Alt key */
    _wasInc;
    _isAltDown;
    _isShiftDown;
    _isCtrlDown;
    gotKeyDown(e) {
        super.gotKeyDown(e);
        if (e.altKey)
            this._isAltDown = true;
        if (e.shiftKey)
            this._isShiftDown = true;
        if (e.ctrlKey)
            this._isCtrlDown = true;
        let v = 0;
        switch (e.key) {
            case "ArrowUp":
                ++v;
                break;
            case "ArrowDown":
                --v;
                break;
            default:
                break;
        }
        if (v) {
            e.preventDefault();
            this.gotIncrement(v);
        }
    }
    /** Called when user tries to increment/decrement value (via ArrowKeys/Mouse/Swipe) */
    gotIncrement(dval) {
        this._wasInc = true;
        if (this._isAltDown) {
            dval *= 0.1;
            if (this.$format.maxDecimal < 1) {
                return; // don't allow decimal increment
            }
        }
        else if (this._isShiftDown)
            dval *= 10;
        else if (this._isCtrlDown)
            dval *= 100;
        const v = +(this.$value ?? 0);
        const hasFloat = this._isAltDown || v % 1 !== 0;
        const next = hasFloat ? mathSumFloat(v, dval) : v + dval;
        const el = this.$refInput;
        const inputType = dval > 0 ? "_inc" : "_dec";
        const data = this.valueToInput(next);
        setTimeout(() => {
            const isPrevented = !el.dispatchEvent(new InputEvent("beforeinput", { inputType, data, bubbles: true, cancelable: true }));
            if (!isPrevented) {
                el.value = data;
                el.dispatchEvent(new InputEvent("input", { inputType, bubbles: true }));
            }
        });
    }
}
customElements.define(tagName, WUPNumberControl);
